/* copy.c */

#include "df3util.h"

static char const * const TU = "copy";

/* globals */
static char *asrc,      /* source name arg, null if default */
            *atgt;      /* target name arg */
static df3Axes aaxes;   /* axes arg */
static Offset aoffs;    /* offset arg */

static df3Axes saxes;   /* source axes arg, if source default */
static size_t sbpv;     /* source bpv arg, if source default */

static df3Handle df3i, df3o;

static bool tgt_l_e;    /* F - sAxes < tAxes.
                         * T - sAxes >= tAxes.
                         */

typedef struct cbData cbData;
struct cbData {
  df3Handle *h;     /* df3o */
  df3Axes i3axes;   /* df3i */
  size_t i3bpv;
  df3Axes axes;     /* arg */
  Offset offs;
};


/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := source axes offset target.
 *     := . axes bpv axes offset target.
 */

static bool procArgs(char *av[]) {
  char buf[PATH_MAX];
  int const rc = getInpSrc(av, &saxes, &sbpv);
  int i;

  if (-1 == rc)
    return false;
  else if (1 == rc) {
    asrc = NULL;
    i = 3;
  }
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    return false;
  }
  else {
    asrc = av[0];
    i = 1;
  }

  if (!(av[i]  &&  av[i+1]  &&  av[i+2])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    return false;
  }

  if (isDefaultArg(av[i]))
    aaxes = (df3Axes)GENORIGIN;
  else if (!getArgCoord(av[i], &aaxes)) {
    EMITEMSG(false, "bad axes value(s)");
    return false;
  }
  else if (!df3_check_axes(aaxes)) {
    EMITEMSG(false, "zero axes value(s)");
    app_lastec = ec_badAxes;
    return false;
  }

  Offset o;
  if (isDefaultArg(av[i+1]))
    o = ZEROOFFS;
  else if (!getArgOffset(av[i+1], &o)) {
    EMITEMSG(false, "bad offset value");
    return false;
  }
  aoffs = o;

  if (!getArgFname(av[i+2], buf)) {
    EMITEMSG(false, "bad or missing target file name");
    return false;
  }
  atgt = av[i+2];

  return true;
}


/* sourceOpenCheck(), targetOpenPrep()
 * -------------------------------------------------------------------------- *
 */

bool sourceOpenCheck(void) {
  df3Error ec;

  if (asrc) {
    if (!mkSourceFname(asrc, false, false, &df3i.fname)) {
      EMITEMSG(false, "source '%s': %s", asrc, appLastError(app_lastec));
      return false;
    }
    else if (ecOk != (ec = df3_open_file(&df3i, true))) {
      EMITEMSG(false, "source '%s': %s",
              trimmedFn(df3i.fname), df3_error_message(ec));
      app_lastec = le2ae(ec);
      return false;
    }
  }
  else {
    df3i.axes = saxes;
    df3i.bpv = sbpv;
  }

  if (!df3_check_axes(aaxes))
    aaxes = df3i.axes;  /* whole source if default */

  if (!haveDiskSpace(aaxes, df3i.bpv)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(atgt), appLastError(app_lastec = ec_diskFull));
    goto label_fail;
  }
  else if (app_diagnose) {
    if (df3i.fname) {
      EMITDMSG(TU, "source handle");
      df3_handle_info(&df3i, app_logfp);
    }
    else
      EMITDMSG(TU, "source handle <stdin>");
  }

  if (!isValidFillval(df3i.bpv))
    goto label_fail;

  if (df3i.axes.x < aaxes.x
          ||  df3i.axes.y < aaxes.y
          ||  df3i.axes.z < aaxes.z) {
    bool junk;
    if (!volsOverlap(aaxes, aoffs, df3i.axes, &junk)) {
      EMITEMSG(false, "no overlap wrt target");
      goto label_fail;
    }
    tgt_l_e = false;
  }
  else {
    bool junk;
    if (!volsOverlap(df3i.axes, aoffs, aaxes, &junk)) {
      EMITEMSG(false, "no overlap wrt source");
      goto label_fail;
    }
    tgt_l_e = true;
  }

  return true;

label_fail:
  if (df3i.fname)
    ec = df3_close_file(&df3i);
// TODO update app_lastec if failed?

  if (app_diagnose) {
    if (df3i.fname)
      EMITDMSG(TU, "source handle closed: %s", df3_handle_error(&df3i, false));
    else
      EMITDMSG(TU, "end source <stdin>");
  }

  return false;
}


static bool targetOpenPrep(void) {
  bool f = false;
  df3Error ec;

  df3o.axes = aaxes;
  df3o.bpv = df3i.bpv;
  if (!mkTargetFname(atgt, false, &df3o.fname))
    EMITEMSG(false, "target '%s': %s", atgt, appLastError(app_lastec));
  else if (ecOk != (ec = df3_create_file(&df3o))) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(df3o.fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3o, app_logfp);
  }

  if (app_fillval  &&  !prepFilled(TU, &df3o)) {
    ec = df3_close_file(&df3o);
// TODO clobbers prepFilled()s.
    app_lastec = le2ae(ec);
  }
  else
    f = true;

  return f;
}


/* closeFiles()
 * -------------------------------------------------------------------------- *
 * TODO what's lost by setting {app_lastec} here?
 * TODO use EMITEMSGs instead + unconditionally?
 */

static bool closeFiles(void) {
  bool f = false;
  df3Error ec;

  if (df3i.fname) {
    if (ecOk != (ec = df3_close_file(&df3i))) {
      if (app_diagnose)
        EMITDMSG(TU, "close source: %s", df3_error_message(ec));
    }
    else if (app_diagnose)
      EMITDMSG(TU, "closed source handle: %s",
              df3_handle_error(&df3i, false));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {  /* redundant, but hey.. */
    (void)fclose(stdin);
    if (app_diagnose)
      EMITDMSG(TU, "closed source <stdin>");
  }

  ec = df3_close_file(&df3o);
  if (ecOk != ec) { 
    if (app_diagnose)
      EMITDMSG(TU, "close target: %s", df3_error_message(ec));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {
    if (app_diagnose)
      EMITDMSG(TU, "closed target handle: %s", df3_handle_error(&df3o, false));
    f = true;
  }

  return f;
}


/* targetGreaterThan(), targetLessEqual()
 * -------------------------------------------------------------------------- *
 * both use either a file loop or a (wrapped) callback to transfer the data.
 * TODO app_warns.
 */

static bool tgtGTcb(void *vp, df3Coord const at, uint32_t const v) {
  cbData *xfer = vp;
  df3Error ec;
  df3Handle * const df3 = xfer->h;
  df3Coord xat = at;

  if (!inpVoxelValid(xfer->i3axes, xfer->i3bpv, at, v))
    return false;
  else if (!atCGTinters(xfer->i3axes, &xat, xfer->offs, df3->axes))
    return true;
  else if (ecOk != (ec = df3_put_voxel(df3, xat, v))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            xat.x, xat.y, xat.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  return true;
}

static bool tgtGTloop(df3Handle *dfi, df3Handle *dfo) {
  bool f = false;
  for (size_t z = 0; dfi->axes.z != z; ++z)
    for (size_t y = 0; dfi->axes.y != y; ++y)
      for (size_t x = 0; dfi->axes.x != x; ++x) {
        df3Coord at, xat;
        df3Error ec;
        uint32_t v;
        at = xat = (df3Coord){x, y, z};
        if (!atCGTinters(dfi->axes, &xat, aoffs, dfo->axes))
          continue;
        else if (ecOk != (ec = df3_get_voxel(dfi, at, &v))) {
          EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        else if (ecOk != (ec = df3_put_voxel(dfo, xat, v))) {
          EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
                  xat.x, xat.y, xat.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }
  f = true;

label_loopfail:
  return f;
}

static bool tgtLEcb(void *vp, df3Coord const at, uint32_t const v) {
  cbData *xfer = vp;
  df3Error ec;
  df3Handle * const df3 = xfer->h;
  df3Coord xat = at;

  if (!inpVoxelValid(xfer->i3axes, xfer->i3bpv, at, v))
    return false;
  else if (!atCLEinters(xfer->i3axes, &xat, xfer->offs, df3->axes))
    return true;
  else if (ecOk != (ec = df3_put_voxel(df3, xat, v))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            xat.x, xat.y, xat.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  return true;
}

static bool tgtLEloop(df3Handle *dfi, df3Handle *dfo) {
  bool f = false;
  for (size_t z = 0; dfi->axes.z != z; ++z)
    for (size_t y = 0; dfi->axes.y != y; ++y)
      for (size_t x = 0; dfi->axes.x != x; ++x) {
        df3Error ec;
        df3Coord at, xat;
        uint32_t v;
        at = xat = (df3Coord){x, y, z};
        if (!atCLEinters(dfi->axes, &xat, aoffs, dfo->axes))
          continue;
        else if (ecOk != (ec = df3_get_voxel(dfi, at, &v))) {
          EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        else if (ecOk != (ec = df3_put_voxel(dfo, xat, v))) {
          EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
                  xat.x, xat.y, xat.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }
  f = true;

label_loopfail:
  return f;
}


static bool targetGreaterThan(void) {
  if (app_diagnose)
    EMITDMSG(TU, "source to larger target");
  bool f;
  if (df3i.fname)
    f = tgtGTloop(&df3i, &df3o);
  else {
    cbData x = {.h = &df3o,
                .i3axes = df3i.axes,
                .i3bpv = df3i.bpv,
                .axes = aaxes,
                .offs = aoffs};
    f = inputGet(NULL, tgtGTcb, &x);
  }
  return f;
}


static bool targetLessEqual(void) {
  if (app_diagnose)
    EMITDMSG(TU, "source to same-size/smaller target");
  bool f;
  if (df3i.fname)
    f = tgtLEloop(&df3i, &df3o);
  else {
    cbData x = {.h = &df3o,
                .i3axes = df3i.axes,
                .i3bpv = df3i.bpv,
                .axes = aaxes,
                .offs = aoffs};
    f = inputGet(NULL, tgtLEcb, &x);
  }
return f;
}


/* ------------------------------------------------------------------------- */

static void copyTidyup(void) {
  FREENULL(df3i.fname);
  FREENULL(df3o.fname);
}

bool copyMain(char *av[]) {
  if (atexit(copyTidyup))
    APPPANIC;

  if (!procArgs(av))
    return false;
  else if (!sourceOpenCheck())
    return false;
  else if (!targetOpenPrep()) {
    (void)df3_close_file(&df3i);
    return false;
  }

  bool const f = (tgt_l_e ? targetLessEqual : targetGreaterThan)();
// TODO update app_lastec if failed?
  if (!closeFiles()) {
    (void)unlink(df3o.fname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed",
              trimmedFn(df3o.fname));
  }

  if (app_verbose  &&  f)
    printf("data copied into %svolume '%s'%s.\n",
            tgt_l_e ? "" : "larger ",
            trimmedFn(df3o.fname),
            app_fillval ? ", filled" : "");

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
