/* dfdr.c */

#include "df3util.h"

#include <argz.h>
#include <strings.h>

#define DFDIDENT  "dfd"
#define DFDLNMAX  512

static char const * const TU = DFDIDENT;


/* infrastructure */
static char const * const ident = DFDIDENT;

enum kIdx {
  kComment = 0, kCyclic, kData, kEndian, kFilename,
  kFormat, kOffset, kType, kXsz, kYsz, kZsz,
  k_k_max
};
typedef enum kIdx kIdx;

static char const * const kword[] = {
  "comment", "cyclic", "data", "endian", "file_name",
  "format", "offset", "type", "x_size", "y_size", "z_size",
  NULL
};


enum tIdx {  /* tests depend on this order */
  tUnknown = 0, tChar,
  tFloat, tFloat8, tFloat16, tFloat32,
  tInt, tInt8, tInt16, tInt32
};
typedef enum tIdx tIdx;

static char const * const tVal[] = {
  "unknown", "char:8bit",
  "float", "float:8bit", "float:16bit", "float:32bit",
  "int", "int:8bit", "int:16bit", "int:32bit",
  NULL
};


/* globals */
static char *asrc;     /* source name */
static char *atgt;     /* target name */

static df3Handle df3o;

static df3Axes svol;   /* kword {x,y,z}_size */
static size_t soffs;   /* kword offset */
static char *sdfn;     /* kword file_name */
static FILE *sfp;      /* EDFD, reuse for file_name + stdin */
static bool sbin;      /* kword format, text=false */
static bool sbig;      /* kword endian, little=false */
static tIdx styp;      /* kword type */
static size_t sbpv;    /* what type equates to */

static bool havekey[k_k_max];  /* ensure unique keywords,
                                * excludes comments.
                                */

static char *scmnts;           /* store comments */
static size_t scmntslen;
static size_t havecomment;     /* number of comments */


static int keywordIndex(char const * const s) {
  for (size_t i = 0; kword[i]; ++i)
    if (!strcasecmp(s, kword[i]))
      return (int)i;
  return -1;
}

static tIdx typeIndex(char const * const s) {
  for (size_t i = 1; tVal[i]; ++i)
    if (!strcasecmp(s, tVal[i]))
      return (tIdx)i;
  return tUnknown;
}

static size_t type2bpv(tIdx const i) {
  size_t r = 0;
  switch (i) {
    case tChar: case tInt8: case tFloat8:
      r = 1; break;
    case tInt: case tInt16: case tFloat: case tFloat16:
      r = 2; break;
    case tInt32: case tFloat32:
      r = 4; break;
    default:
      APPPANIC;
  }
  return r;
}

static void prettyEdfd(bool const inln) {
  int const w = 16;
  char vdata[DFDLNMAX];
  strcpy(vdata, sdfn ? trimmedFn(sdfn) : "<stdin>");
  argz_stringify(scmnts, scmntslen, CHRNLN);
  EMITDMSG(TU, "source description file.\n"
               "%*s: %s\n"
               "%*s: %zu, %zu, %zu\n"
               "%*s: %s (%zu)\n"
               "%*s: %s\n"
               "%*s: %s\n"
               "%*s: %zu\n"
               "%*s: %s\n"
               "%*s: %s",
          w, "name", trimmedFn(asrc),
          w, "volume (x,y,z)", svol.x, svol.y, svol.z,
          w, "type + bpv", tVal[(size_t)styp], sbpv,
          w, "format", sbin ? "binary" : "text",
          w, "endian", sbin ? sbig ? "big" : "little" : "n/a",
          w, "offset", soffs,
          w, "data source", inln ? "inline" : vdata,
          w, "comments", havecomment ? scmnts : "n/a" );
  return;
}


static bool intTextCvt(char const * const ln,
                       uint32_t const lim,
                       uint32_t *val) {
  uint32_t tmp = 0;
  if (!getArgValue(ln, &tmp))
    return false;
  else if (lim < tmp) {
    app_lastec = ec_badArgument;
    return false;
  }
  *val = tmp;
  return true;
}
static bool linei2i32(char const * const a, uint32_t *v) {
  return intTextCvt(a, UINT32_MAX, v);
}
static bool linei2i16(char const * const a, uint32_t *v) {
  return intTextCvt(a, UINT16_MAX, v);
}
static bool linei2i8(char const * const a, uint32_t *v) {
  return intTextCvt(a, UINT8_MAX, v);
}

static bool floatTextCvt(char const * const ln,
                         uint32_t const lim,
                         uint32_t *val) {
  double const u = 1. / (1 + (double)lim);
  float n = 0.0;
  if (!getArgFloat(ln, &n))
    return false;
  *val = (uint32_t)((double)n / u) - (1 == n ? 1 : 0);
  return true;
}
static bool linef2i32(char const * const a, uint32_t *v) {
  return floatTextCvt(a, UINT32_MAX, v);
}
static bool linef2i16(char const * const a, uint32_t *v) {
  return floatTextCvt(a, UINT16_MAX, v);
}
static bool linef2i8(char const * const a, uint32_t *v) {
  return floatTextCvt(a, UINT8_MAX, v);
}


/* getAsrc()
 * -------------------------------------------------------------------------- *
 * if fname absolute check that, else check as is and if not found prepend
 * workdir.  no dice? error.  allocate name as asrc.  return.
 */

static bool viableSrc(char const * const fn) {
  if (!statFileExists(fn))
    return false;
  bool junk = false;
  return statFilePerms(fn, &junk);
}

static bool getAsrc(char const * const fn) {
  char const fmt[] = "source '%s': %s";
  char buf[PATH_MAX] = {CHRNUL};

  if (PATH_MAX < strlen(fn)) {
    EMITEMSG(false, fmt, trimmedFn(fn),
            appLastError(app_lastec = ec_badFilename));
    return false;
  }
  else if (PTHSEP == *fn) {  /* absolute */
    if (!viableSrc(fn)) {
      EMITEMSG(false, fmt, trimmedFn(fn),
              appLastError(app_lastec = ec_noFile));
      return false;
    }
    strcpy(buf, fn);
  }
  else {  /* relative */
    if (viableSrc(fn))
      strcpy(buf, fn);
    else {
      char buf2[PATH_MAX];
      if (PATH_MAX == snprintf(buf2, PATH_MAX, "%s%c%s",
              app_workdir ? app_workdir : ".", PTHSEP, fn)) {
        EMITEMSG(false, fmt, trimmedFn(buf2),
                appLastError(app_lastec = ec_badFilename));
        return false;
      }
      else if (!viableSrc(buf2)) {
        EMITEMSG(false, fmt, trimmedFn(buf2),
                appLastError(app_lastec = ec_noFile));
        return false;
      }
      strcpy(buf, buf2);
    }
  }

  asrc = strdup(buf);
  if (!asrc)
    APPPANIC;

  return true;
}


/* parseEdfd()
 * -------------------------------------------------------------------------- *
 * parse an extended density file description and initialise the corresponding
 * globals.  magic numbers (cf regex.c): kbuf longest = 9, vbuf longest = 499.
 * TODO rewrite.
 */

static bool checkHaveAll(void) {
  char const * const fmt = "missing keyword [%s] in '%s'";
  bool f = false;
  if (!havekey[kFormat]) {
    EMITEMSG(false, fmt, kword[kFormat], trimmedFn(asrc));
    app_lastec = ec_badArgument;
  }
  else if (!havekey[kType]) {
    EMITEMSG(false, fmt, kword[kType], trimmedFn(asrc));
    app_lastec = ec_badArgument;
  }
  else if (sbin  &&  !havekey[kEndian]) {
    EMITEMSG(false, fmt, kword[kEndian], trimmedFn(asrc));
    app_lastec = ec_badArgument;
  }
  else if (!(havekey[kXsz]  &&  havekey[kYsz]  &&  havekey[kZsz])) {
    EMITEMSG(false, "missing keyword(s) [?_size] in '%s'",
            trimmedFn(asrc));
    app_lastec = ec_badArgument;
  }
  else if (!df3_check_axes(svol)) {
    EMITEMSG(false, "invalid value(s) in [?_size] in '%s'", trimmedFn(asrc));
    app_lastec = ec_badAxes;
  }
  else if (!(3 < svol.x  &&  3 < svol.y)) {
    EMITEMSG(false, "[x_size] or [y_size] is less than '4' in '%s'",
            trimmedFn(asrc));
    app_lastec = ec_badAxes;
  }
  else if (!haveDiskSpace(svol, sbpv)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(atgt), appLastError(app_lastec = ec_diskFull));
    return false;
  }
  else
    f = true;
  return f;
}

static bool parseLine(char const * const line, int const lnum, bool *kdata) {
  char kbuf[10] = {CHRNUL}, vbuf[500] = {CHRNUL};
  int const rc = getEdfdKV(line, kbuf, vbuf);
  if (!rc) return false;

  int const ti = keywordIndex(kbuf);
  if (-1 == ti) {
    EMITEMSG(false, "unknown key in '%s' line %d", trimmedFn(asrc), lnum);
    app_lastec = ec_badArgument;
    return false;
  }
  else if (havekey[ti]) {
    EMITEMSG(false, "duplicate key in '%s' line %d", trimmedFn(asrc), lnum);
    app_lastec = ec_badArgument;
    return false;
  }

  char const * const fmt = "bad/unknown value in '%s' line %d";
  size_t const i = (size_t)ti;
  if (1 == rc) {
    if (kFilename == (kIdx)i)
      havekey[i] = true;
    else if (kData == (kIdx)i) {
      havekey[i] = true;
      if (!checkHaveAll())
        return false;
      *kdata = true;
    }
    else {
      EMITEMSG(false, "missing value for [%s], in '%s' line %d",
              kword[i], trimmedFn(asrc), lnum);
      app_lastec = ec_badArgument;
      return false;
    }
  }
  else
    switch ((kIdx)i) {  /* nb some braced */
      case kComment:
        if (argz_add(&scmnts, &scmntslen, vbuf))
          APPPANIC;
        havecomment += 1;
        break;
      case kCyclic:
        if (!strcasecmp("off", vbuf))
          ;  /* noop */
        else if (!strcasecmp("on", vbuf)) {
          EMITEMSG(false, "keyword [cyclic] is not yet implemented");
          app_lastec = ec_badArgument;
          return false;
        }
        else {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        havekey[i] = true;
        break;
      case kEndian:
        if (!strcasecmp("big", vbuf))
          sbig = true;
        else if (!strcasecmp("little", vbuf))
          sbig = false;
        else {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        havekey[i] = true;
        break;
      case kFilename: {
        char fname[DFDLNMAX] = {CHRNUL};
        if (!getArgFname(vbuf, fname)) {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        else if (!viableSrc(fname)) {
          char s[DFDLNMAX]; 
          strcpy(s, trimmedFn(fname));
          EMITEMSG(false, "missing file '%s', in '%s' line %d",
                  s, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        else if (!(sdfn = strdup(fname)))
          APPPANIC;
        havekey[i] = true;
        break;
      }
      case kFormat:
        if (!strcasecmp("binary", vbuf))
          sbin = true;
        else if (!strcasecmp("text", vbuf))
          sbin = false;
        else {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        havekey[i] = true;
        break;
      case kOffset: {
        uint32_t t = 0;
        if (!getArgValue(vbuf, &t)) {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        soffs = (size_t)t;
        havekey[i] = true;
        break;
      }
      case kType: {
        tIdx const t = typeIndex(vbuf);
        if (tUnknown == t) {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        styp = t;
        sbpv = type2bpv(t);
        havekey[i] = true;
        break;
      }
      case kXsz: {
        uint32_t t = 0;
        if (!getArgValue(vbuf, &t)) {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        svol.x = (size_t)t;
        havekey[i] = true;
        break;
      }
      case kYsz: {
        uint32_t t = 0;
        if (!getArgValue(vbuf, &t)) {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        svol.y = (size_t)t;
        havekey[i] = true;
        break;
      }
      case kZsz: {
        uint32_t t = 0;
        if (!getArgValue(vbuf, &t)) {
          EMITEMSG(false, fmt, trimmedFn(asrc), lnum);
          app_lastec = ec_badArgument;
          return false;
        }
        svol.z = (size_t)t;
        havekey[i] = true;
        break;
      }
      default:
        EMITEMSG(false, "stray value for [%s], in '%s' line %d",
                kword[i], trimmedFn(asrc), lnum);
        return false;
    }

  return true;
}

static bool inlineDataTest(bool *isinline) {
  off_t const pos1 = ftello(sfp);
  if (-1 == pos1) APPPANIC;
  if (-1 == fseeko(sfp, 0, SEEK_END))
    APPPANIC;
  off_t const pos2 = ftello(sfp);
  if (-1 == pos2) APPPANIC;


  char const * const msg = "too many voxel data sources";
  bool f = false;
  if (2 >= (pos2 - pos1)) {  /* allow stray newline(s) */
    if (!(sdfn  ||  itDfd == app_inptype)) {
      EMITEMSG(false, "no voxel data");
      app_lastec = ec_badArgument;
      return false;
    }
    else if (sdfn  &&  itDfd == app_inptype) {
      EMITEMSG(false, "%s", msg);
      app_lastec = ec_badArgument;
      return false;
    }
    f = false;
  }
  else {
// TODO times width of type.
    int64_t const n = (int64_t)(soffs + svol.x * svol.y * svol.z);
    if (n > (int64_t)(pos2 - pos1)) {
      EMITEMSG(false, "not enough voxel data");
      app_lastec = ec_badArgument;
      return false;
    }
    else if (sdfn  ||  itDfd == app_inptype) {
      EMITEMSG(false, "%s", msg);
      app_lastec = ec_badArgument;
      return false;
    }
    else if (sbin) {
      EMITEMSG(false, "inline data must be \"[format] text\"");
      app_lastec = ec_badArgument;
      return false;
    }
    f = true;
  }

  if (f)
    if (-1 == fseeko(sfp, pos1, SEEK_SET))
      APPPANIC;

  *isinline = f;

  return true;
}

static bool parseEdfd(void) {

  char line[DFDLNMAX] = {CHRNUL};

  if (!(sfp = fopen(asrc, "rb")))
    APPPANIC;
  else if (!fgets(line, DFDLNMAX, sfp)) {
    int const e = errno;
    app_lastec = ec_sysError;
    EMITEMSG(false, "reading source '%s': %s",
            trimmedFn(asrc), strerror(e));
    return false;
  }
  else if (4 != strlen(line)  ||  strncasecmp(ident, line, 3)) {
    EMITEMSG(false, "missing '%s' identifier in '%s' line 1",
            ident, trimmedFn(asrc));
    app_lastec = ec_noFile;
    return false;
  }

  int lnum = 1;
  while (fgets(line, DFDLNMAX, sfp)) {
    bool endloop = false;
    lnum += 1;
    if (!parseLine(line, lnum, &endloop))
      return false;
    else if (endloop)
      break;
  }

  bool isinline = false;
  if (ferror(sfp))
    APPPANIC;
  else if (!inlineDataTest(&isinline))
    return false;
  else if (app_diagnose)
    prettyEdfd(isinline);

  if (!isinline) {
    if (fclose(sfp)) {
      int const e = errno;
      app_lastec = ec_sysError;
      sfp = NULL;
      EMITEMSG(false, "closing source description '%s': %s",
              trimmedFn(asrc), strerror(e));
      return false;
    }
    else if (app_diagnose)
      EMITDMSG(TU, "closed source description file");

    if (itDfd == app_inptype)
      sfp = stdin;
    else {
      sfp = fopen(sdfn, "rb");
      if (!sfp) {
        int const e = errno;
        app_lastec = ec_sysError;
        EMITEMSG(false, "opening source '%s': %s",
                trimmedFn(sdfn), strerror(e));
        return false;
      }
    }
    if (app_diagnose)
      EMITDMSG(TU, "%s data source '%s'",
             sdfn ? "opened" : "using",
             sdfn ? trimmedFn(sdfn) : "<stdin>");
  }
  else if (app_diagnose)
    EMITDMSG(TU, "data source description: using inline data");

  return true;
}


/* openTarget()
 * -------------------------------------------------------------------------- *
 * create target, fill if required.  true if all went well.
 */

static bool openTarget(void) {
  df3o = DF3_INITHANDLE;  /* redundant, but hey.. */
  df3o.fname = atgt;
  df3o.axes = svol;
  df3o.bpv = sbpv;

  df3Error const ec = df3_create_file(&df3o);
  if (ecOk != ec) {
    EMITEMSG(false, "create target '%s': %s",
            trimmedFn(df3o.fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3o, app_logfp);
  }

  if (!prepFilled(TU, &df3o)) {
    (void)df3_close_file(&df3o);
    return false;
  }

  return true;
}


/* procBinary(), procText()
 * -------------------------------------------------------------------------- *
 * the respective handlers for the processing of the voxel data.
 */

static bool procBinary(void) {
EMITEMSG(false, "binary source data not yet implemented");
return false;
}


static inline void stripNln(char *s) {
  size_t const n = strlen(s) - 1;
  if (CHRNLN == s[n])
    s[n] = CHRNUL;
}

static bool procText(void) {
  char line[DFDLNMAX];

  if (soffs) {
    size_t n = soffs;
    while (fgets(line, DFDLNMAX, sfp)  &&  --n)
      ;  /* noop */
    if (ferror(sfp))
      APPPANIC;
    else if (feof(sfp)) {
      app_lastec = ec_badArgument;
      EMITEMSG(false, "data source unexpected EOF");
      return false;
    }
  }

  /* set conversion func */
  bool const t = tChar < styp  &&  tInt > styp;
  bool (*fn)(char const * const, uint32_t *);
  switch (sbpv) {
    case 4:  fn = t ? linef2i32 : linei2i32; break;
    case 2:  fn = t ? linef2i16 : linei2i16; break;
    case 1:  fn = t ? linef2i8 : linei2i8; break;
    default: APPPANIC;
  }

  bool f = false;
  for (size_t z = 0; svol.z != z; ++z)
    for (size_t y = 0; svol.y != y; ++y)
      for (size_t x = 0; svol.x != x; ++x) {
        df3Axes at = {x, y, z};
        df3Error ec;
        uint32_t v;
        if (!fgets(line, DFDLNMAX, sfp)) {
          int const e = errno;
          app_lastec = ec_sysError;
          EMITEMSG(false, "reading data source: %s", strerror(e));
          goto label_loopfail;
        }
        stripNln(line);
        if (!fn(line, &v)) {
          EMITEMSG(false, "conversion failed for <%zu,%zu,%zu>",
                  at.x, at.y, at.z);
          goto label_loopfail;
        }
        else if (ecOk != (ec = df3_put_voxel(&df3o, at, v))) {
          EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }

  f = true;

label_loopfail:
  return f;
}


/* ------------------------------------------------------------------------- */

static void dfdTidyup(void) {
  FREENULL(scmnts);
  FREENULL(sdfn);
  if (sfp)
    (void)fclose(sfp);
  FREENULL(asrc);
  FREENULL(atgt);
}

bool dfdMain(char *av[]) {
  if (atexit(dfdTidyup))
    APPPANIC;
  else if (!av[1]) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    return false;
  }
  else if (!mkTargetFname(av[1], false, &atgt)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(av[1]), appLastError(app_lastec));
    return false;
  }
  else if (!getAsrc(av[0]))
    return false;
  else if (!parseEdfd())
    return false;
  else if (!openTarget())
    return false;

  bool f = (sbin ? procBinary : procText)();

  df3Error ec = df3_close_file(&df3o);
  if (!f  ||  ec != ecOk) {
    (void)unlink(df3o.fname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(df3o.fname));
    f = false;
  }
  else if (app_diagnose)
    EMITDMSG(TU, "closed target handle: %s", df3_handle_error(&df3o, false));

  if (fclose(sfp)) {
    int const e = errno;
    EMITEMSG(false, "closing data source '%s': %s", sdfn ? trimmedFn(sdfn) :
                    itDfd == app_inptype ? "<stdin>" : trimmedFn(asrc),
            strerror(e));
  }
  else if (app_diagnose)
    EMITDMSG(TU, "closed data source '%s'", sdfn ? trimmedFn(sdfn) :
                    itDfd == app_inptype ? "<stdin>" : trimmedFn(asrc));
  sfp = NULL;

  if (f  &&  app_verbose) {
    char buf[DFDLNMAX];
    strcpy(buf, trimmedFn(df3o.fname));
    printf("created target '%s', axes <%zu,%zu,%zu> bpv %zu,\n"
           "    from edfd '%s'.\n",
            buf, svol.x, svol.y, svol.z, sbpv, trimmedFn(asrc));
  }

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
