/* stat.c */

#include "df3util.h"

#include <sys/stat.h>
#include <sys/statvfs.h>


/* setupGrp()
 * -------------------------------------------------------------------------- *
 * cf 'info libc' section 29.2 ff.
 */

static uid_t euid;
static gid_t egid;
static gid_t *sgroups;
static size_t nsg;

static void tidyupGrp(void) {
  FREENULL(sgroups);
  return;
} 

void setupGrp(void) {
  int const n = getgroups(0, NULL);
  if (-1 == n)
    APPPANIC;
  else if (!n)
    return;

  nsg = (size_t)n;
  gid_t *tmp = malloc(sizeof(gid_t) * nsg);
  if (!tmp)
    APPPANIC;
  else if (-1 == getgroups(n, tmp))
    APPPANIC;
  else if (atexit(tidyupGrp))
    APPPANIC;
  sgroups = tmp;

  euid = geteuid();
  egid = getegid();
  return;
}

/* check presence */
static bool isInGroups(gid_t const g) {
  for (size_t i = 0; i != nsg; ++i)
    if (g == sgroups[i])
      return true;
  return false;
}


/* filter 'real' errors */
static bool realErr(int const e, bool const f) {
  if (EFAULT == e  ||  ELOOP == e  ||  ENOMEM == e  ||  EOVERFLOW == e) {
    EMITEMSG(false, "stat%s: %s", f ? "vfs" : "", strerror(e));
    exit(EXIT_FAILURE);
  }
  return false;
}


bool statFileExists(char const * const fn) {
  errno = 0;
  struct stat sb;
  int const rc = stat(fn, &sb);
  if (rc) {
    if (ENOENT == errno)
      return false;
    return realErr(errno, false);
  }

  return S_ISREG(sb.st_mode);
}


bool statDirectoryPerms(char const * const dn) {
  errno = 0;
  struct stat sb;
  int const rc = stat(dn, &sb);
  if (rc) {
    if (ENOENT == errno)
      return false;
    return realErr(errno, false);
  }
  else if (!S_ISDIR(sb.st_mode))
    return false;

  mode_t perms;
  if (euid == sb.st_uid)
    perms = S_IRWXU;
  else if (egid == sb.st_gid  ||  isInGroups(sb.st_gid))
    perms = S_IRWXG;
  else
    perms = S_IRWXO;

  return (perms & sb.st_mode) == perms;
}


bool statFilePerms(char const * const fn, bool *iswr) {
  errno = 0;
  struct stat sb;
  int const rc = stat(fn, &sb);
  if (rc) {
    /* no file is ok */
    if (ENOENT == errno)
      return true;
    return realErr(errno, false);
  }
  else if (!S_ISREG(sb.st_mode))
    return false;

  mode_t rperm, wperm;
  if (euid == sb.st_uid)
    rperm = S_IRUSR, wperm = S_IWUSR;
  else if (egid == sb.st_gid  ||  isInGroups(sb.st_gid))
    rperm = S_IRGRP, wperm = S_IWGRP;
  else
    rperm = S_IROTH, wperm = S_IWOTH;

  if (wperm & sb.st_mode)
    *iswr = true;

  return rperm & sb.st_mode;
}


bool haveDiskSpace(df3Axes const a, size_t const b) {
  errno = 0;
  struct statvfs vfb;
  int const rc = statvfs(app_workdir ? app_workdir : ".", &vfb);
  if (-1 == rc) {
    if (EIO == errno  ||  ENOSYS == errno) {
      EMITEMSG(false, "statvfs: %s", strerror(errno));
      exit(EXIT_FAILURE);
    }
    return realErr(errno, true);
  }
  else if (!vfb.f_bavail)
    return false;
  /* include 8k for overheads (header,inode) */
  uint64_t const n = (uint64_t)(a.x * a.y * a.z * b + 8192) / vfb.f_bsize;
  return vfb.f_bavail > n;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
